/* DDS Sine Generator / 3 phase PWM Inverter mit Atmega32u4 ver.0.1 Alfa
Timer4 generates the 15625 Hz Clock Interrupt */

#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <Wire.h>
#include <LiquidCrystal_I2C.h>
 
byte temp_cel_deg[8] = {0b01100,0b10010,0b10010,0b01100,0b00000,0b00000,0b00000,0b00000};

LiquidCrystal_I2C lcd(0x3F,16,2);  // set the LCD address to 0x3F for a 16 chars and 2 line display

// table of 8 bit 256 sine values / 1/2 sine period / stored in flash memory
const byte PROGMEM sine256[] = 
{0,3,6,9,12,15,18,21,25,28,31,34,37,40,43,46,49,52,55,59,62,65,68,71,74,77,80,83,86,89,91,94,97,100,103,106,109,
112,114,117,120,123,125,128,131,134,136,139,141,144,147,149,152,154,157,159,162,164,166,169,171,173,176,178,180,
182,185,187,189,191,193,195,197,199,201,203,205,207,208,210,212,214,215,217,219,220,222,223,225,226,228,229,230,
232,233,234,236,237,238,239,240,241,242,243,244,245,246,247,247,248,249,249,250,251,251,252,252,253,253,253,254,
254,254,255,255,255,255,255,255,255,255,255,255,255,254,254,254,253,253,253,252,252,251,251,250,249,249,248,247,
247,246,245,244,243,242,241,240,239,238,237,236,234,233,232,230,229,228,226,225,223,222,220,219,217,215,214,212,
210,208,207,205,203,201,199,197,195,193,191,189,187,185,182,180,178,176,173,171,169,166,164,162,159,157,154,152,
149,147,144,141,139,136,134,131,128,125,123,120,117,114,112,109,106,103,100,97,94,91,89,86,83,80,77,74,71,68,65,
62,59,55,52,49,46,43,40,37,34,31,28,25,21,18,15,12,9,6,3};

#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

volatile byte sine_table_indexA;
volatile byte sine_table_indexA_Prev;
volatile byte sine_table_indexB;
volatile byte sine_table_indexB_Prev;
volatile byte sine_table_indexC;
volatile byte sine_table_indexC_Prev;

volatile word phase_accumulator;
volatile int tuning_word;
volatile long tuning_word_soft;

volatile int tuning_word_ADC;
int Temp_Monitor;
float Temp_Monitor_Float;
word LCD_Buttons_ADC;

volatile word PWM_Soft_A;
volatile word PWM_Soft_B;
volatile word PWM_Soft_C;
volatile word Amp;
volatile boolean phaseA_sign;
volatile boolean phaseB_sign;
volatile boolean phaseC_sign;

volatile byte phaseShiftA; 
volatile byte phaseShiftB; 
volatile byte phaseShiftC;

volatile word PhaseCount;
volatile byte StateVar;
volatile int Revers;

volatile byte LCD_Button;
char myStr[10]; 
int SpeedIncrease = 1024;

int nsamples = 10;
float average = 0;

void setup()
{
Serial.begin(256000);   // connect to the serial port

lcd.init();
lcd.begin(16,2);                     
lcd.setCursor(14, 0);
lcd.print("Hz");
lcd.createChar(0, temp_cel_deg); 
lcd.setCursor(14, 1);
lcd.print(char(0));
lcd.print("C");
lcd.home();

pinMode(4,  OUTPUT); // sets PD4 as output / Program_Fault OUT
pinMode(7,  INPUT);  // sets PE6/AIN0 as input / Device_Fault IN
pinMode(13, OUTPUT); // pin13 / OC4A  / PWM output phase A
pinMode(5,  OUTPUT); // pin5  / #OC4A / PWM output phase inv A
pinMode(10, OUTPUT); // pin10 / OC4B  / PWM output phase B
pinMode(9,  OUTPUT); // pin9  / #OC4B / PWM output phase inv B
pinMode(6,  OUTPUT); // pin6  / OC4D  / PWM output phase C
pinMode(12, OUTPUT); // pin12 / #OC4D / PWM output phase inv C

//**************************************************************************

digitalWrite(4, LOW); //////////////////////////////////////////////////////////
 
cbi (TIMSK0, TOIE0); // disable Timer0 Interrupt
cbi (TIMSK1, TOIE1); // disable Timer1 Interrupt
cbi (TIMSK4, TOIE4); // disable Timer4 Interrupt

cbi (PLLFRQ, PDIV0);        // PLL Lock Frequency 96MHz 
sbi (PLLFRQ, PDIV1); 
cbi (PLLFRQ, PDIV2); 
sbi (PLLFRQ, PDIV3); 
cbi (PLLFRQ, PLLTM0);       // PLL Postcaler for Timer4 / PLL/1,5 = 64MHz
sbi (PLLFRQ, PLLTM1); 
sbi (PLLFRQ, PLLUSB);       // PLL Postcaler for USB Peripheral / PLL/2 = 48MHz
cbi (PLLFRQ, PINMUX);       // Use Primary System Clock for PLL

sbi (PLLCSR, PINDIV);       // Using a 16MHz clock source for PLL
sbi (PLLCSR, PLLE);         // PLL enable

cbi (CLKPR, CLKPS0);        // Clock Prescaler Select to 4
sbi (CLKPR, CLKPS1);
cbi (CLKPR, CLKPS2);
cbi (CLKPR, CLKPS3);

TCCR4A = 0;
TCCR4B = 0;
TCCR4C = 0;
TCCR4D = 0;

// Timer4 Clock Prescaler to : 1 / Timer4 Phase and Frequency Correct PWM Mode, 9-bit / TOP OCR4C (0x01FF)

sbi (TCCR4A, COM4A0); // Clear OC4A/OC4B on Compare Match when up-counting. Set on Compare Match when down-counting.
cbi (TCCR4A, COM4A1); // Enable complementary Compare Output mode
sbi (TCCR4A, COM4B0); 
cbi (TCCR4A, COM4B1);
sbi (TCCR4A, PWM4A);  // Pulse Width Modulator A Enable
sbi (TCCR4A, PWM4B);  // Pulse Width Modulator B Enable

sbi (TCCR4B, CS40);   // Timer/Counter4 Prescaler Select / CK/4
sbi (TCCR4B, CS41);
cbi (TCCR4B, CS42);
cbi (TCCR4B, CS43);
cbi (TCCR4B, DTPS40); // Dead Time Prescaler Bits / no division
cbi (TCCR4B, DTPS41);

sbi (TCCR4C, COM4D0); // Clear OC4D on Compare Match when up-counting. Set on Compare Match when down-counting.
cbi (TCCR4C, COM4D1); // Enable complementary Compare Output mode
sbi (TCCR4C, PWM4D);  // Pulse Width Modulator D Enable

sbi (TCCR4D, WGM40);  // Phase and Frequency Correct PWM / TOP - OCR4C
cbi (TCCR4D, WGM41);

DT4 = 0x00;           // Timer/Counter4 Dead Time Value = 0

                      // TIFR4 - TC4 Interrupt Flag Register
cbi (TIFR4, TOV4);    // TOV - Timer/Counter4, Overflow Flag
cbi (TIFR4, OCF4A);   // Timer/Counter4, Output Compare A Match Flag
cbi (TIFR4, OCF4B);   // Timer/Counter4, Output Compare B Match Flag
cbi (TIFR4, OCF4D);   // Timer/Counter4, Output Compare C Match Flag

TC4H = 0x01;          // Timer/Counter4 PWM TOP 0x01FF / 9-bits resolution
OCR4C = 0xFF;        
//sbi (TIMSK4, TOIE4);  // Timer/Counter4, Overflow Interrupt Enable
}

void loop()
{ 
//   Serial.write(PWM_Soft_A);
//   Serial.write(PWM_Soft_B);
//   Serial.write(PWM_Soft_C);
//   Serial.write(sine_table_indexA);
//   Serial.write(sine_table_indexB);
//   Serial.write(sine_table_indexC);
//tuning_word_ADC = 1;
   
tuning_word_ADC = analogRead(A4);
Temp_Monitor    = analogRead(A3);
LCD_Buttons_ADC = analogRead(A0);                   //read the buttons
if (LCD_Buttons_ADC > 1000) LCD_Button = 0;         //buttonNONE 
else if (LCD_Buttons_ADC < 50)   LCD_Button = 0;    //buttonRIGHT;  
else if (LCD_Buttons_ADC < 195)  LCD_Button = 1;    //buttonUP; 
else if (LCD_Buttons_ADC < 380)  LCD_Button = 2;    //buttonDOWN; 
else if (LCD_Buttons_ADC < 555)  LCD_Button = 0;    //buttonLEFT; 
else if (LCD_Buttons_ADC < 790)  LCD_Button = 5;    //buttonSELECT;   
else LCD_Button = 0;

//lcd.setCursor(0, 0);
//lcd.print(StateVar);

lcd.setCursor(0, 0);
sprintf (myStr, "%4d", tuning_word);
lcd.print(myStr);

lcd.setCursor(4, 0);
sprintf (myStr, "%4d", tuning_word_ADC);
lcd.print(myStr);

//lcd.setCursor(0, 1);
//sprintf (myStr, "%3d", phaseShiftA);
//lcd.print(myStr);
//sprintf (myStr, "%3d", phaseShiftB);
//lcd.print(myStr);
//sprintf (myStr, "%3d", phaseShiftC);
//lcd.print(myStr);

lcd.setCursor(10, 0);
float frieq;
frieq = tuning_word/8.372224; // 511*256/15625 Hz
dtostrf(frieq, 3, 1, myStr);
lcd.print(myStr);

#define B 4092 // B-коэффициент
#define SERIAL_R 3600 // сопротивление последовательного резистора, 3,6 кОм
#define THERMISTOR_R 85000 // номинальное сопротивления термистора, 85 кОм
#define NOMINAL_T 25 // номинальная температура (при которой TR = 85 кОм)

//    average = 0;
//  for (int i = 0; i <= nsamples; i++) 
//  {
//    average = average + Temp_Monitor;
//  } 
//    average = average / nsamples;  
    float steinhart;
    float tr = 1023.0 / Temp_Monitor - 1;
    tr = SERIAL_R / tr;
    steinhart = tr / THERMISTOR_R; // (R/Ro)
    steinhart = log(steinhart); // ln(R/Ro)
    steinhart /= B; // 1/B * ln(R/Ro)
    steinhart += 1.0 / (NOMINAL_T + 273.15); // + (1/To)
    steinhart = 1.0 / steinhart; // Invert
    steinhart -= 273.15; 

char str_temp[4];        
dtostrf(steinhart, 2, 1, str_temp);
sprintf(myStr,"%s", str_temp);
lcd.setCursor(10, 1);
lcd.print(myStr);

if (Revers == 0)
  {
if (LCD_Button == 5)                          // Stop
  {StateVar = 0; Revers = 0;}
else if (LCD_Button == 1 && StateVar == 0)    // Forward
  {StateVar = 1; Revers = 0; sbi (TIMSK4, TOIE4);}
else if (LCD_Button == 2 && StateVar == 0)    // Backward
  {StateVar = 2; Revers = 0; sbi (TIMSK4, TOIE4);}
else if (LCD_Button == 2 && StateVar == 1)    // Forward to Backward 
  {StateVar = 0; Revers =-1;}    
else if (LCD_Button == 1 && StateVar == 2)    // Backward to Forward
  {StateVar = 0; Revers = 1;}
  }
 else if (Revers == -1 && tuning_word == 0) {StateVar = 2; Revers = 0; Index_Clear(); sbi (TIMSK4, TOIE4);}
 else if (Revers ==  1 && tuning_word == 0) {StateVar = 1; Revers = 0; Index_Clear(); sbi (TIMSK4, TOIE4);}
}

ISR(TIMER4_OVF_vect)                         // Timer4 Interrupt
{  
  if (StateVar == 0) {tuning_word_ADC = 0;}    
  if (tuning_word == 0 && StateVar == 0) 
  {
  cbi (TIMSK4, TOIE4); Index_Clear(); 
  }
  
   if      (tuning_word < tuning_word_ADC)
    {tuning_word_soft += SpeedIncrease; tuning_word = tuning_word_soft >> 17;}
   else if (tuning_word > tuning_word_ADC)
    {
    tuning_word_soft -= SpeedIncrease; tuning_word = tuning_word_soft >> 17; 
    if (tuning_word_soft <= 0) {tuning_word_soft = 0; tuning_word = 0;}
    }
 
  if (tuning_word < 8)
    {Amp = 7;}
  else if (tuning_word >= 8 && tuning_word <= 418)
    {Amp = tuning_word >> 2;}
  else
   {Amp = 127;}
//   Amp = 64;
      
  switch(StateVar)
    {
      case 0:
        PWM(phaseShiftA, phaseShiftB, phaseShiftC);
      break;
      case 1: 
        phaseShiftA = 0; phaseShiftB = 170; phaseShiftC = 85;
        PWM(phaseShiftA, phaseShiftB, phaseShiftC);    
      break;
      case 2:
        phaseShiftA = 170; phaseShiftB = 0; phaseShiftC = 85;
        PWM(phaseShiftA, phaseShiftB, phaseShiftC);         
      break;
    }
}
      
void PWM(byte phaseShiftA, byte phaseShiftB, byte phaseShiftC)
{    
phase_accumulator += tuning_word;                                                          // Update phase accumulator and extract the sine table index from it
sine_table_indexA_Prev = sine_table_indexA;
sine_table_indexA = highByte(phase_accumulator) + phaseShiftA;                             // Use upper 8 bits phase accumulator as index phase A
sine_table_indexB_Prev = sine_table_indexB;
sine_table_indexB = highByte(phase_accumulator) + phaseShiftB;                             // Use upper 8 bits phase accumulator as index phase B
sine_table_indexC_Prev = sine_table_indexC;
sine_table_indexC = highByte(phase_accumulator) + phaseShiftC;                             // Use upper 8 bits phase accumulator as index phase C

 if (sine_table_indexA_Prev > sine_table_indexA)
  {phaseA_sign = not phaseA_sign;}
 if (sine_table_indexB_Prev > sine_table_indexB)
  {phaseB_sign = not phaseB_sign;}
 if (sine_table_indexC_Prev > sine_table_indexC)
  {phaseC_sign = not phaseC_sign;}

  if (phaseA_sign == true)
    {PWM_Soft_A = 256 + (((pgm_read_byte_near(sine256 + sine_table_indexA))*Amp) >> 7);}     // Set current amplitude value for the 1/2 sine wave phase A being constructed.       
  else
    {PWM_Soft_A = 255 - (((pgm_read_byte_near(sine256 + sine_table_indexA))*Amp) >> 7);}     // Set current amplitude value for the 1/2 sine wave phase A being constructed.  
  if (phaseB_sign == true)
    {PWM_Soft_B = 256 + (((pgm_read_byte_near(sine256 + sine_table_indexB))*Amp) >> 7);}     // Set current amplitude value for the 1/2 sine wave phase B being constructed.   
  else
    {PWM_Soft_B = 255 - (((pgm_read_byte_near(sine256 + sine_table_indexB))*Amp) >> 7);}     // Set current amplitude value for the 1/2 sine wave phase B being constructed.   
  if (phaseC_sign == false)
    {PWM_Soft_C = 256 + (((pgm_read_byte_near(sine256 + sine_table_indexC))*Amp) >> 7);}     // Set current amplitude value for the 1/2 sine wave phase C being constructed.     
  else
    {PWM_Soft_C = 255 - (((pgm_read_byte_near(sine256 + sine_table_indexC))*Amp) >> 7);}     // Set current amplitude value for the 1/2 sine wave phase C being constructed.     
    
      sbi (TCCR4E, TLOCK4);               // Synchronous update
      TC4H = highByte(PWM_Soft_A); 
      OCR4A = lowByte(PWM_Soft_A);    
      TC4H = highByte(PWM_Soft_B); 
      OCR4B = lowByte(PWM_Soft_B);
      TC4H = highByte(PWM_Soft_C); 
      OCR4D = lowByte(PWM_Soft_C);   
      cbi (TCCR4E, TLOCK4);
}
void Index_Clear() 
{
tuning_word_soft = 0; 
tuning_word = 0; 
phase_accumulator = 0;
phaseShiftA = 0; 
phaseShiftB = 0; 
phaseShiftC = 0;  
sine_table_indexA_Prev = 0;
sine_table_indexB_Prev = 0;
sine_table_indexC_Prev = 0;
phaseA_sign = false;
phaseB_sign = false;
phaseC_sign = false;
}
